# 可视化的方式设置mesh和material的自定义属性

import bpy
import time
from . import toolFunctions


# Properties
class MeshProperties(bpy.types.PropertyGroup):
    use_default_alphaindex: bpy.props.BoolProperty(
        name = '使用默认 AlphaIndex',
        description = '使用默认 AlphaIndex',
        default = True
    )
    alphaindex: bpy.props.IntProperty(
        name = 'AlphaIndex',
        description = 'AlphaIndex',
        default = 1
    )
    pickable: bpy.props.BoolProperty(
        name = 'pickable',
        description = '关闭时该物体会在事件中被忽略',
        default = True
    )
    reflection: bpy.props.BoolProperty(
        name = '在水中产生倒影',
        description = '开启时物体会在水中产生倒影',
        default = False
    )
    castshadow: bpy.props.BoolProperty(
        name = '产生阴影',
        description = '开启时物体会产生阴影',
        default = True
    )
    receiveshadows: bpy.props.BoolProperty(
        name = '接收阴影',
        description = '开启时物体会接收阴影',
        default = True
    )
    is_tree: bpy.props.BoolProperty(
        name = '树',
        description = '该物体是树。',
        default = False
    )
    is_water: bpy.props.BoolProperty(
        name = '水',
        description = '该物体是水。',
        default = False
    )


class MaterialProperties(bpy.types.PropertyGroup):
    needdepthprepass: bpy.props.BoolProperty(
        name = 'NeedDepthPrePass',
        description = 'NeedDepthPrePass',
        default = False
    )
    use_glowlayer: bpy.props.BoolProperty(
        name = '使用发光层',
        description = '使用发光层',
        default = False
    )
    glowlayer: bpy.props.IntProperty(
        name = '发光层',
        description = '使用的发光层的编号',
        default = 1,
        min = 1
    )



# Panel
class MeshPanel(bpy.types.Panel):
    bl_label = '扩展属性'
    bl_idname = 'MESH_PT_EXTRA_PROPERTIES'
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = 'data'

    def draw(self, context):
        props = context.mesh.DSProperties_Properties
        self.layout.prop(props, 'use_default_alphaindex')
        if not props.use_default_alphaindex:
            self.layout.prop(props, 'alphaindex')
        self.layout.prop(props, 'pickable')
        self.layout.prop(props, 'reflection')
        self.layout.prop(props, 'castshadow')
        self.layout.prop(props, 'receiveshadows')
        self.layout.prop(props, 'is_tree')
        self.layout.prop(props, 'is_water')
        self.layout.operator('scene.cal_water_reflectioned_trees', text = '计算/更新倒影树')

class MaterialPanel(bpy.types.Panel):
    bl_label = '扩展属性'
    bl_idname = 'MATERIAL_PT_EXTRA_PROPERTIES'
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = 'material'

    def draw(self, context):
        if context.material is None:
            return
        props = context.material.DSProperties_Properties
        self.layout.prop(props, 'needdepthprepass')
        self.layout.prop(props, 'use_glowlayer')
        if props.use_glowlayer:
            self.layout.prop(props, 'glowlayer')



# Operator
class Operator(bpy.types.Operator):
    bl_label = '计算/更新倒影树'
    bl_idname = 'scene.cal_water_reflectioned_trees'

    def execute(self, context):
        start = time.time()
        waterObjects = []
        treeObjects = []

        for object in bpy.data.objects:
            if 'minPow2Len' in object:
                del object['minPow2Len']
            if 'DSProperties_Properties' not in object.data or object.users_collection[0].hide_render:
                continue
            props = object.data.DSProperties_Properties
            if props.is_water:
                waterObjects.append(object)
            if props.is_tree:
                treeObjects.append(object)

        for treeObject in treeObjects:
            p = treeObject.location
            minPow2Len = float('inf')
            minp1 = []
            minp2 = []
            for waterObject in waterObjects:
                for edgeKey in waterObject.data.edge_keys:
                    p1 = waterObject.matrix_world @ waterObject.data.vertices[edgeKey[0]].co
                    p2 = waterObject.matrix_world @ waterObject.data.vertices[edgeKey[1]].co
                    p1p = toolFunctions.sub(p, p1)
                    p2p = toolFunctions.sub(p, p2)
                    p2p1 = toolFunctions.sub(p1, p2)
                    p1p2 = toolFunctions.sub(p2, p1)
                    dotP1 = toolFunctions.dot(p1p, p1p2)
                    dotP2 = toolFunctions.dot(p2p, p2p1)
                    pow2Len = 0

                    if dotP1 <= 0:
                        # P1P
                        pow2Len = toolFunctions.pow2len(p1p)
                    elif dotP2 <= 0:
                        # P2P
                        pow2Len = toolFunctions.pow2len(p2p)
                    else:
                        # P to P1P2
                        pow2Len = toolFunctions.pow2len(p1p) - dotP1 * dotP1 / toolFunctions.pow2len(p2p1)

                    if pow2Len <= minPow2Len:
                        minp1 = p1
                        minp2 = p2
                    minPow2Len = min(minPow2Len, pow2Len)

            treeObject['minPow2Len'] = minPow2Len

        print('Time cost:', time.time() - start, 's')
        return { 'FINISHED' }



# Register / Unregister
def register():
    bpy.utils.register_class(MeshProperties)
    bpy.utils.register_class(MeshPanel)
    bpy.utils.register_class(MaterialProperties)
    bpy.utils.register_class(MaterialPanel)
    bpy.utils.register_class(Operator)
    bpy.types.Mesh.DSProperties_Properties = bpy.props.PointerProperty(type = MeshProperties)
    bpy.types.Material.DSProperties_Properties = bpy.props.PointerProperty(type = MaterialProperties)

def unregister():
    bpy.utils.unregister_class(MeshProperties)
    bpy.utils.unregister_class(MeshPanel)
    bpy.utils.unregister_class(MaterialProperties)
    bpy.utils.unregister_class(MaterialPanel)
    bpy.utils.unregister_class(Operator)
    del bpy.types.Mesh.DSProperties_Properties
    del bpy.types.Material.DSProperties_Properties



def gather_material_hook(user_extension, gltf2_material, blender_material, export_settings):
    props = blender_material.DSProperties_Properties
    if not gltf2_material.extras:
        gltf2_material.extras = {}

    if props.needdepthprepass:
        gltf2_material.extras['needDepthPrePass'] = 'yes'
    if props.use_glowlayer:
        gltf2_material.extras['glowLayer'] = props.glowlayer

def gather_mesh_hook(user_extension, gltf2_mesh, blender_mesh, blender_object, vertex_groups, modifiers, skip_filter, material_names, export_settings):
    props = blender_mesh.DSProperties_Properties

    if not gltf2_mesh.extras:
        gltf2_mesh.extras = {}
    if not props.use_default_alphaindex:
        gltf2_mesh.extras['alphaIndex'] = props.alphaindex
    gltf2_mesh.extras['pickable'] = props.pickable
    gltf2_mesh.extras['reflection'] = props.reflection
    gltf2_mesh.extras['receiveShadows'] = props.receiveshadows
    gltf2_mesh.extras['castShadow'] = props.castshadow

def gather_node_hook(user_extension, gltf2_node, blender_object, export_settings):
    if 'minPow2Len' in blender_object:
        if not gltf2_node.extras:
            gltf2_node.extras = {}
        gltf2_node.extras['is_reflection_tree'] = blender_object['minPow2Len'] <= 100