# 导出材质上的动画

import bpy

extension_is_required = False
glTF_extension_name = "DS_materials_animation"

class MaterialsAnimationProperties(bpy.types.PropertyGroup):
    enabled: bpy.props.BoolProperty(
        name = '材质动画',
        description = '导出材质上的动画',
        default = True
    )

def register():
    bpy.utils.register_class(MaterialsAnimationProperties)
    bpy.types.Scene.MaterialsAnimationProperties = bpy.props.PointerProperty(type = MaterialsAnimationProperties)


def unregister():
    bpy.utils.unregister_class(MaterialsAnimationProperties)
    del bpy.types.Scene.MaterialsAnimationProperties


def pannel_draw(panel, context):
    layout = panel.layout
    layout.prop(bpy.context.scene.MaterialsAnimationProperties, 'enabled')


def parent_data_path(data_path):
    index = data_path.rfind('.')
    if index == -1:
        return data_path
    return data_path[:index]

def get_value(blender_fcurve, frame):
    return blender_fcurve.evaluate(frame)

def next_nodes(node):
    nodes = {}
    output_idx = 0
    for output in node.outputs:
        for link in output.links:
            if output_idx not in nodes:
                nodes[output_idx] = []
            nodes[output_idx].append(link.to_node)
        output_idx = output_idx + 1
    return nodes


def get_action_param_paths(animated_node, animated_property, params):
    ret = None
    if animated_node.name == 'MOZ_lightmap settings':
        return [['lightmapTexture', 'level']]
    if animated_node.name == 'Principled BSDF':
        if animated_property.name in params:
            if ret is None:
                ret = []
            animated_property_type = 'Default'
            if animated_property.type == 'RGBA':
                animated_property_type = 'Color'
            ret.append([animated_property.name, animated_property_type])

    animated_node_next_nodes = next_nodes(animated_node)
    if len(animated_node_next_nodes) == 0:
        return None
    for output_idx in animated_node_next_nodes:
        next_node_idx = -1
        for next_node in animated_node_next_nodes[output_idx]:
            next_node_idx = next_node_idx + 1
            if next_node is None:
                continue
            if next_node.type == "BSDF_PRINCIPLED":
                type = animated_node.outputs[output_idx].links[next_node_idx].to_socket.name
                animation_type = animated_property.name
                if ret is None:
                    ret = []
                ret.append([type, animation_type])
            next_next_nodes = next_nodes(next_node)
            for next_output_idx in next_next_nodes:
                next_next_node_idx = -1
                for next_next_node in next_next_nodes[next_output_idx]:
                    next_next_node_idx = next_next_node_idx + 1
                    if next_next_node is None:
                        continue
                    if next_next_node.type == "BSDF_PRINCIPLED":
                        texture_type = next_node.outputs[next_output_idx].links[next_next_node_idx].to_socket.name + " Texture"
                        animation_type = animated_property.name
                        if ret is None:
                            ret = []
                        ret.append([texture_type, animation_type])
    return ret

def gather_extension_to_ref(animated_property, action_param_paths, blender_fcurves, extension_container):
    # action_param_paths 可能为 None 代表这个 blender_action 没有被用到
    if action_param_paths is None:
        return
    # 遍历帧 确定开始结束时间
    start_frame = float("inf")
    end_frame = 0
    for fcurve in blender_fcurves:
        for keyframe_point in fcurve.keyframe_points:
            start_frame = min(start_frame, keyframe_point.co[0])
            end_frame = max(start_frame, keyframe_point.co[0])

    # TODO 这边因为光照贴图的 intensity 属性并不能连接值 所以暂时只能特殊写出来 如果有需要考虑统一处理这类属性
    if animated_property.name == 'MOZ_lightmap settings':
        default_value = animated_property.intensity
    else:
        default_value = animated_property.default_value
    frame_arr = []
    value_arr = []
    vector_len = 1
    if type(default_value) != float:
        vector_len = len(default_value)
    frame = start_frame
    while(frame <= end_frame):
        for i in range(0, vector_len):
            if vector_len == 1:
                value = default_value
            else:
                value = default_value[i]
            for blender_fcurve in blender_fcurves:
                if blender_fcurve.array_index == i:
                    value = get_value(blender_fcurve, frame)
                    break
            value_arr.append(value)
        frame_arr.append(frame)
        frame += 1.0

    for action_param_path in action_param_paths:
        param1 = action_param_path[0]
        param2 = action_param_path[1]
        if param1 not in extension_container:
            extension_container[param1] = {}

        try:
            param1.index('Texture')
        except:
            # do nothing
            param1
        else:
            # blender 和 babylonjs 对于 uv 中 v 的处理不太一样  所以对 voffset 取反
            if param2 == 'Location':
                for idx in range(len(value_arr)):
                    if (idx - 1) % 3 == 0:
                        value_arr[idx] = -value_arr[idx]

        extension = {}
        extension['fps'] = bpy.context.scene.render.fps
        extension['frames'] = frame_arr
        extension['values'] = value_arr
        extension['value_len'] = vector_len

        extension_container[param1][param2] = extension


def gather_material_hook(user_extension, gltf2_material, blender_material, export_settings):
    if not bpy.context.scene.MaterialsAnimationProperties.enabled: return

    if blender_material.node_tree.animation_data is None: return

    blender_action = blender_material.node_tree.animation_data.action
    blender_fcurves = blender_action.fcurves
    if len(blender_fcurves) == 0:
        return

    params = ["Emission Strength", 'Emission', "Alpha"]
    extension_container = {}

    animated_node = None
    animated_property = None
    data_path = None
    node_fcurves = []
    blender_fcurves_count = len(blender_fcurves)
    for fcurve_idx in range(blender_fcurves_count):
        blender_fcurve = blender_fcurves[fcurve_idx]

        if blender_fcurve.data_path == data_path:
            node_fcurves.append(blender_fcurve)
        # 这个 fcurve 与前一个 fcurve 改的不是同一个属性（对于 blender 来说的属性）
        # 这时需要生成上一个属性的 animation
        if blender_fcurve.data_path != data_path:
            if len(node_fcurves) > 0:
                action_param_paths = get_action_param_paths(animated_node, animated_property, params)
                gather_extension_to_ref(animated_property, action_param_paths, node_fcurves, extension_container)
                node_fcurves = []
            data_path = blender_fcurve.data_path
            animated_node = eval('blender_material.node_tree.' + parent_data_path(parent_data_path(data_path)))
            animated_property = eval('blender_material.node_tree.' + parent_data_path(data_path))
            node_fcurves.append(blender_fcurve)

        if fcurve_idx == blender_fcurves_count - 1:
            action_param_paths = get_action_param_paths(animated_node, animated_property, params)
            gather_extension_to_ref(animated_property, action_param_paths, node_fcurves, extension_container)

    gltf2_material.extensions[glTF_extension_name] = user_extension.Extension(
        name = glTF_extension_name,
        extension = extension_container,
        required = extension_is_required
    )